"""
A module for asymmetric's callback schemas.
"""


CALLBACK_OBJECT_SCHEMA = {
    "callback_url_header": {
        "required": False,
        "type": str,
    },
    "callback_method_header": {
        "required": False,
        "type": str,
    },
    "custom_callback_key_header": {
        "required": False,
        "type": str,
    },
}


CALLBACK_OBJECT_DEFAULTS = {
    "callback_url_header": "Asymmetric-Callback-URL",
    "callback_method_header": "Asymmetric-Callback-Method",
    "custom_callback_key_header": "Asymmetric-Custom-Callback-Key",
}


CALLBACK_OBJECT_METADATA = {
    "callback_url_header": {
        "required": True,
        "description": "URL to the API to send the function response data.",
    },
    "callback_method_header": {
        "required": False,
        "description": (
            "HTTP method to use when making the callback request. Defaults to POST."
        ),
    },
    "custom_callback_key_header": {
        "required": False,
        "description": (
            "Key to wrap the function output around. By default, "
            "the output won't be wrapped."
        ),
    },
}



"""
A module for containing the singleton metaclass for the _Asymmetric class.
"""

from typing import Any


class AsymmetricSingleton(type):

    """
    Singleton metaclass to limit the existance of the asymmetric object to one.
    """

    def __call__(cls, *args: Any, **kwargs: Any) -> Any:
        """Defines the class creation method."""
        if not hasattr(cls, "asymmetric_instance"):  # Object does not exist
            cls.asymmetric_instance = super().__call__(*args, **kwargs)  # Create object
        return cls.asymmetric_instance  # Return asymmetric object



"""
A module to hold some CLI utilities.
"""

import json
import os
import sys
from importlib import import_module
from traceback import print_exception
from types import ModuleType

from asymmetric.core import _Asymmetric
from asymmetric.errors import AppImportError
from asymmetric.helpers import humanize
from asymmetric.openapi.core import get_openapi


def document_openapi(module: str, filename: str) -> None:
    """
    Gets the asymmetric object and then calls the OpenAPI spec generator method.
    """
    asymmetric_object = get_asymmetric_object(module)
    docs = get_openapi(asymmetric_object, f"{humanize(module)} API")
    with open(filename, "w") as docs_file:
        json.dump(docs, docs_file, indent=2)


def get_main_module(module_name: str) -> ModuleType:
    """
    Imports the module :module_name and returns it. This method is strongly
    inspired in gunicorn's own import_app method and uvicorn's main method.
    """
    try:
        # Add current directory to path
        sys.path.insert(0, ".")
        return import_module(module_name)
    except ImportError as error:
        # If the user wrote module.py instead of just module
        if module_name.endswith(".py") and os.path.exists(module_name):
            actual_name = module_name.rsplit(".", 1)[0]
            error_str = f"Module {module_name} not found. Did you mean {actual_name}?"
            raise ImportError(error_str) from error
        raise


def get_asymmetric_object(module_name: str, debug: bool = True) -> _Asymmetric:
    """
    Imports the module :module_name and the tries to find the
    asymmetric object. This method is strongly inspired in gunicorn's own
    import_app method and uvicorn's main method.
    """
    # Import the module
    module = get_main_module(module_name)

    # Get the asymmetric object
    try:
        return getattr(module, "asymmetric")
    except AttributeError as error:
        if debug:
            print_exception(*sys.exc_info())
        error_str = f"Failed to find the asymmetric object in {module_name}."
        raise AppImportError(error_str) from error



"""
A module for asymmetric's callback utilities.
"""

import json
from typing import Any, Dict, Union

from asymmetric.callbacks.callback_object import (
    CALLBACK_OBJECT_DEFAULTS,
    CALLBACK_OBJECT_SCHEMA,
)
from asymmetric.errors import InvalidCallbackObjectError
from asymmetric.utils import valid_plain_dict


def valid_callback_data(callback: Union[Dict[str, Any], bool]) -> bool:
    """
    Returns a boolean indicating if the callback object is a valid json
    conforming to the json schema.
    """
    if isinstance(callback, bool):
        return True
    return valid_plain_dict(callback, CALLBACK_OBJECT_SCHEMA)


def validate_callback_data(callback: Union[Dict[str, Any], bool]) -> None:
    """Raises an error if the callback data is not valid."""
    if not valid_callback_data(callback):
        raise InvalidCallbackObjectError(
            "Invalid callback object:\n"
            + json.dumps(callback, indent=2, sort_keys=False, ensure_ascii=False)
        )


def get_header_finders(callback: Union[Dict[str, Any], bool]) -> Dict[str, str]:
    """Gets the header finders. Returns the default ones if it doesn't find some."""
    if isinstance(callback, bool):
        # Callback attribute is True
        return CALLBACK_OBJECT_DEFAULTS
    # Callback attribute is an object, get callback attribute finders
    return {
        k: callback[k] if k in callback else default_value
        for k, default_value in CALLBACK_OBJECT_DEFAULTS.items()
    }



"""
A module for the interactive documentation renderers.
"""

from asymmetric.constants import OPENAPI_SPEC_ROUTE


def get_redoc_html(title: str) -> str:
    """Gets the HTML template for rendering the ReDoc interactive documentation."""
    redoc_script = "https://cdn.jsdelivr.net/npm/redoc@next/bundles/redoc.standalone.js"
    google_fonts = (
        "https://fonts.googleapis.com/css?family=Montserrat:300,400,700|"
        "Roboto:300,400,700"
    )

    html = f"""
    <!DOCTYPE html>
    <html>
        <head>
            <title>{title}</title>
            <!-- needed for adaptive design -->
            <meta charset="utf-8"/>
            <meta
                name="viewport"
                content="width=device-width, initial-scale=1"
            >
            <link href="{google_fonts}" rel="stylesheet">
            <style>
            body {{
                margin: 0;
                padding: 0;
            }}
            </style>
        </head>
            <body>
            <redoc spec-url="{OPENAPI_SPEC_ROUTE}">
            </redoc>
            <script src="{redoc_script}"></script>
        </body>
    </html>
    """
    return html


def get_swagger_html(title: str) -> str:
    """Gets the HTML template for rendering the Swagger interactive documentation."""
    swagger_js_url = (
        "https://cdn.jsdelivr.net/npm/swagger-ui-dist@3.30.0/swagger-ui-bundle.js"
    )
    swagger_css_url = (
        "https://cdn.jsdelivr.net/npm/swagger-ui-dist@3.30.0/swagger-ui.css"
    )

    html = f"""
    <!DOCTYPE html>
    <html>
        <head>
            <link type="text/css" rel="stylesheet" href="{swagger_css_url}">
            <title>{title}</title>
        </head>
        <body>
            <div id="swagger-ui"></div>
            <script src="{swagger_js_url}"></script>
            <!-- `SwaggerUIBundle` is now available on the page -->
            <script>
                const ui = SwaggerUIBundle({{
                    url: '{OPENAPI_SPEC_ROUTE}',
                    dom_id: '#swagger-ui',
                    presets: [
                    SwaggerUIBundle.presets.apis,
                    SwaggerUIBundle.SwaggerUIStandalonePreset
                    ],
                    layout: "BaseLayout",
                    deepLinking: true,
                    showExtensions: true,
                    showCommonExtensions: true
                }})
            </script>
        </body>
    </html>
    """
    return html



"""
A module for asymmetric's core callback logic.
"""

import asyncio
from typing import Any, Callable, Dict, Optional, Union

import httpx
from starlette.datastructures import Headers
from starlette.responses import JSONResponse

from asymmetric.callbacks.utils import get_header_finders, validate_callback_data
from asymmetric.constants import HTTP_METHODS
from asymmetric.errors import InvalidCallbackHeadersError, InvalidCallbackObjectError
from asymmetric.loggers import log
from asymmetric.utils import generic_call


class CallbackClient:

    """
    A class to abstract the callback logic.
    """

    def __init__(
        self, function: Callable[..., Any], callback: Union[Dict[str, Any], bool]
    ) -> None:
        self.__function = function
        self.__attribute_finders = self.__prepare_and_validate_finders(callback)
        self.__headers = Headers()
        self.__params: Dict[str, str] = {}
        self.__invalid_callback_object = ""

    @property
    def url(self) -> str:
        """Returns the callback URL if it was included, or None."""
        location = self.__attribute_finders.get("callback_url_header")
        return self.__headers.get(location, "") if location is not None else ""

    @property
    def http_method(self) -> str:
        """Returns the callback HTTP method if it was included, or POST."""
        location = self.__attribute_finders.get("callback_method_header", "")
        # POST is the default HTTP method
        http_method = self.__headers[location] if location in self.__headers else "POST"
        return http_method.upper()

    @property
    def custom_key(self) -> Optional[str]:
        """Returns the custom callback key container if it was included, or None."""
        location = self.__attribute_finders.get("custom_callback_key_header", "")
        if location in self.__headers:
            return self.__headers.get(location)
        return None

    def handle_callback(self, headers: Headers, params: Dict[str, Any]) -> JSONResponse:
        """
        Validates that the callback data from the request is correct
        and delegates the main function call. Returns a JSON response.
        """
        if self.__invalid_callback_object:
            # Callback object was defective, server should have stopped
            return JSONResponse(
                {"message": self.__invalid_callback_object}, status_code=500
            )

        try:
            # Set headers and params
            self.__headers = headers
            self.__params = params

            # Validate callback data
            self.__validate_callback_url()
            self.__validate_callback_http_method()

            # Delegate function
            asyncio.ensure_future(self.__callback_call())

            return JSONResponse({}, status_code=202)
        except InvalidCallbackHeadersError as error:
            return JSONResponse({"message": str(error)}, status_code=422)

    @staticmethod
    def __prepare_and_validate_finders(
        callback: Union[Dict[str, Any], bool]
    ) -> Dict[str, str]:
        """
        Collects and returns the callback data finders and validates
        that they are correct on decoration-time.
        """
        try:
            validate_callback_data(callback)
            return get_header_finders(callback)
        except InvalidCallbackObjectError as error:
            log(str(error), level="critical")
            raise InvalidCallbackObjectError(error) from error

    def __validate_callback_url(self) -> None:
        """
        Validates that the callback URL within the request
        header is valid.
        """
        if not self.url:
            raise InvalidCallbackHeadersError("Invalid callback URL")

    def __validate_callback_http_method(self) -> None:
        """
        Validates that the callback HTTP method within the request
        header is valid.
        """
        if self.http_method.lower() not in HTTP_METHODS:
            raise InvalidCallbackHeadersError("Invalid callback HTTP method")

    async def __callback_call(self) -> None:
        """
        Executes the function and makes the request to the callback endpoint.
        """
        try:
            response = await generic_call(self.__function, self.__params)
            if self.custom_key is not None:
                response = {self.custom_key: response}

            async with httpx.AsyncClient() as client:
                await client.request(
                    self.http_method,
                    self.url,
                    json=response,
                )
        except Exception as error:
            message = "Error while executing the delegated method: "
            message += str(error)
            log(message, level="warn")



"""
A module for every helper of the OpenAPI documentation generator.
"""

from typing import Any


def type_to_string(type_obj: Any) -> str:
    """Given a python type, return its JSON schema string counterpart."""
    try:
        type_str = type_obj.__name__
        if type_str == "str":
            return "string"
        if type_str == "float":
            return "number"
        if type_str == "int":
            return "integer"
        if type_str == "bool":
            return "boolean"
        if type_str == "NoneType":
            return "null"
        if type_str == "list":
            return "array"
        return "object"
    except Exception:
        return "object"



"""
Module to hold the openapi documentation constants.
"""


ANY_TYPE = [
    {"type": "string"},
    {"type": "number"},
    {"type": "integer"},
    {"type": "boolean"},
    {"type": "array"},
    {"type": "object"},
]



"""
Module to hold the openapi documentation creation utilities.
"""

import functools
from inspect import FullArgSpec, getfullargspec
from typing import TYPE_CHECKING, Any, Dict, List

from asymmetric.callbacks.callback_object import CALLBACK_OBJECT_METADATA
from asymmetric.callbacks.utils import get_header_finders
from asymmetric.endpoints import Endpoint
from asymmetric.openapi.constants import ANY_TYPE
from asymmetric.openapi.helpers import type_to_string

if TYPE_CHECKING:  # pragma: nocover
    from asymmetric.core import _Asymmetric


def get_parameters_amount(params: FullArgSpec) -> Dict[str, int]:
    """
    Gets a params object retrieved from a function with getfullargspec and returns
    a dictionary with the amount of default and not default arguments.
    """
    parameters_amount = len(params.args)
    defaults_amount = 0 if params.defaults is None else len(params.defaults)
    no_defaults_amount = parameters_amount - defaults_amount
    return {"no_defaults": no_defaults_amount, "defaults": defaults_amount}


def get_no_defaults_schema(params: FullArgSpec) -> Dict[str, Any]:
    """Gets the parameters without a default option of the method's params."""
    no_defaults = get_parameters_amount(params)["no_defaults"]
    schema: Dict[str, Any] = {}
    for index in range(no_defaults):
        param = params.args[index]
        param_label = "type" if param in params.annotations else "oneOf"
        param_type: Any = (
            type_to_string(params.annotations[param])
            if param in params.annotations
            else ANY_TYPE
        )
        schema[param] = {param_label: param_type}
    return schema


def get_defaults_schema(params: FullArgSpec) -> Dict[str, Any]:
    """Gets the parameters with a default option of the method's params."""
    params_amount = get_parameters_amount(params)
    defaults = params_amount["defaults"]
    offset = params_amount["no_defaults"]
    schema: Dict[str, Any] = {}
    for index in range(defaults):
        param = params.args[offset + index]
        default_value = params.defaults[index]  # type: ignore
        param_label = "type" if param in params.annotations else "oneOf"
        param_type: Any = (
            type_to_string(params.annotations[param])
            if param in params.annotations
            else ANY_TYPE
        )
        schema[param] = {param_label: param_type, "default": default_value}
    return schema


def get_openapi_endpoint_body_schema(endpoint: Endpoint) -> Dict[str, Any]:
    """Assembles the OpenAPI schema for the endpoint body."""
    params = getfullargspec(endpoint.function)
    no_defaults_schema = get_no_defaults_schema(params)
    defaults_schema = get_defaults_schema(params)
    return {
        "type": "object",
        "properties": {**no_defaults_schema, **defaults_schema},
        "additionalProperties": params.varkw is not None,
    }


def get_openapi_endpoint_headers_schema(endpoint: Endpoint) -> List[Dict[str, Any]]:
    """Assembles the OpenAPI schema for the endpoint headers."""
    if not endpoint.callback:
        return []
    header_data = get_header_finders(endpoint.callback)
    return [
        {
            "in": "header",
            "name": header,
            "schema": {
                "type": "string",
            },
            "required": CALLBACK_OBJECT_METADATA[key]["required"],
            "description": CALLBACK_OBJECT_METADATA[key]["description"],
        }
        for key, header in header_data.items()
    ]


def get_openapi_endpoint_responses_schema(endpoint: Endpoint) -> Dict[str, Any]:
    """Assembles the OpenAPI schema for the endpoint responses."""
    params = getfullargspec(endpoint.function)
    response_type = "AcceptedOperation" if endpoint.callback else "SuccesfulOperation"
    responses: Dict[str, Any] = {
        f"{endpoint.response_code}": {
            "$ref": f"#/components/responses/{response_type}"
        },
        "500": {"$ref": "#/components/responses/InternalError"},
    }
    if endpoint.callback is False and "return" in params.annotations:
        responses[f"{endpoint.response_code}"]["content"] = {
            "application/json": {
                "schema": {"type": type_to_string(params.annotations["return"])}
            }
        }
    return responses


def get_openapi_endpoint_schema(route_dict: Dict[str, Endpoint]) -> Dict[str, Any]:
    """
    Generate the OpenAPI documentation for dictionary of endpoints for a specific route.
    """
    route_schema = {}
    for http_method, endpoint in route_dict.items():
        headers_schema = get_openapi_endpoint_headers_schema(endpoint)
        body_schema = get_openapi_endpoint_body_schema(endpoint)
        responses_schema = get_openapi_endpoint_responses_schema(endpoint)
        route_schema[http_method.lower()] = {
            "description": endpoint.docstring,
            "parameters": headers_schema,
            "responses": responses_schema,
        }
        has_properties = bool(body_schema["properties"])
        has_body = has_properties or bool(body_schema["additionalProperties"])
        if has_body:
            route_schema[http_method.lower()]["requestBody"] = {
                "required": has_properties,
                "content": {"application/json": {"schema": body_schema}},
            }
    return route_schema


def get_openapi_components() -> Dict[str, Any]:
    """Gets the OpenAPI components."""
    return {
        "responses": {
            "SuccesfulOperation": {
                "description": "Operation completed successfully",
            },
            "AcceptedOperation": {
                "description": "Operation delegated successfully",
            },
            "InternalError": {
                "description": "Unexpected internal error (API method "
                "failed, probably due to a missuse of the "
                "underlying function)",
            },
        },
    }


def get_openapi(
    asymmetric_object: "_Asymmetric",
    title: str,
    version: str = "0.0.1",
    openapi_version: str = "3.0.3",
) -> Dict[str, Any]:
    """
    Gets the OpenAPI spec of every endpoint and assembles it into a
    JSON formatted object.
    """
    # pylint: disable=W0212
    endpoints = asymmetric_object._Asymmetric__endpoints.endpoints
    return {
        "openapi": openapi_version,
        "info": {"title": title, "version": version},
        "paths": functools.reduce(
            lambda x, y: {**x, **y},
            [
                {route: get_openapi_endpoint_schema(route_dictionary)}
                for route, route_dictionary in endpoints.items()
            ],
            {},
        ),
        "components": get_openapi_components(),
    }



"""
A module for containing the endpoint logic of asymmetric.
"""

from inspect import getdoc
from typing import Any, Callable, Dict, List, Optional, Union

from asymmetric.errors import DuplicatedEndpointError


class Endpoint:

    """
    Class to encapsulate an endpoint.
    """

    def __init__(
        self,
        route: str,
        method: str,
        function: Callable[..., Any],
        decorated_function: Callable[..., Any],
        callback: Union[Dict[str, Any], bool] = False,
        response_code: int = 200,
    ) -> None:
        self.__route: str = route
        self.__method: str = method
        self.__function: Callable[..., Any] = function
        self.__decorated_function: Callable[..., Any] = decorated_function
        self.__callback = callback
        self.__response_code: int = response_code

    @property
    def route(self) -> str:
        """Returns the route of the endpoint."""
        return self.__route

    @property
    def method(self) -> str:
        """Returns the HTTP method of the endpoint."""
        return self.__method

    @property
    def response_code(self) -> int:
        """Returns the response code of the endpoint on a request."""
        return self.__response_code if not self.__callback else 202

    @property
    def callback(self) -> Union[Dict[str, Any], bool]:
        """Returns the response code of the endpoint on a request."""
        return self.__callback

    @property
    def function(self) -> Callable[..., Any]:
        """Returns the raw function of the endpoint."""
        return self.__function

    @property
    def docstring(self) -> str:
        """Returns the docstring content of the endpoint function."""
        docstring = getdoc(self.__function)
        return docstring or "No description provided."


class Endpoints:  # pylint: disable=R0903

    """
    Class to encapsulate the endpoints logic.
    """

    def __init__(self) -> None:
        self.__endpoints: Dict[str, Dict[str, Endpoint]] = {}

    @property
    def endpoints(self) -> Dict[str, Dict[str, Endpoint]]:
        """Returns the endpoints dict."""
        return self.__endpoints

    def add_endpoints(
        self,
        route: str,
        methods: List[str],
        function: Callable[..., Any],
        decorated_function: Callable[..., Any],
        callback: Union[Dict[str, Any], bool] = False,
        response_code: int = 200,
    ) -> None:
        """
        Adds an endpoint for every method specified.
        """
        for method in methods:
            self.__add_endpoint(
                route,
                method,
                function,
                decorated_function,
                callback=callback,
                response_code=response_code,
            )

    def __add_endpoint(
        self,
        route: str,
        method: str,
        function: Callable[..., Any],
        decorated_function: Callable[..., Any],
        callback: Union[Dict[str, Any], bool] = False,
        response_code: int = 200,
    ) -> None:
        """
        Checks if the desired endpoint does not exist. If it exists,
        it raises an error. Otherwise, it creates the endpoint
        and stores it.
        """
        if self.__get_endpoint(route, method) is not None:
            message = f"Endpoint '{route}' with HTTP method '{method.upper()}' "
            message += "was defined twice."
            raise DuplicatedEndpointError(message)

        endpoint = Endpoint(
            route,
            method,
            function,
            decorated_function,
            callback=callback,
            response_code=response_code,
        )

        route_dictionary = self.__get_route(route)
        if route_dictionary is None:
            route_dictionary = self.__create_route(route)
        route_dictionary[method] = endpoint

    def __get_route(self, route: str) -> Optional[Dict[str, Endpoint]]:
        """
        Returns the route dictionary containing each method
        as a key. If the route has not been used, returns None.
        """
        if route not in self.__endpoints:
            return None
        return self.__endpoints[route]

    def __create_route(self, route: str) -> Dict[str, Endpoint]:
        """Creates a route and returns it."""
        self.__endpoints[route] = {}
        return self.__endpoints[route]

    def __get_endpoint(self, route: str, method: str) -> Optional[Endpoint]:
        """
        Returns the endpoint object for a specific route/method
        combination. If said combination does not exist, returns None.
        """
        route_data = self.__get_route(route)
        if route_data is None:
            return None
        if method not in route_data:
            return None
        return route_data[method]



"""
A module to route the CLI traffic.
"""

import sys
from argparse import ArgumentParser, _SubParsersAction
from typing import Any

import asymmetric
from asymmetric.cli.utils import document_openapi


def dispatcher(*args: Any, **kwargs: Any) -> None:
    """
    Main CLI method, recieves the command line action and dispatches it to
    the corresponding method.
    """
    parser = generate_parser()
    parsed_args = parser.parse_args(*args, **kwargs)

    try:
        if parsed_args.action == "docs":
            document_openapi(parsed_args.module, parsed_args.filename)
    except AttributeError:
        print("An argument is required for the asymmetric command.")
        parser.print_help()
        sys.exit(1)


def generate_parser() -> ArgumentParser:
    """Generates the CLI parser for the module."""
    # Create parser
    parser = ArgumentParser(description="Command line interface tool for asymmetric.")

    # Add version command
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"asymmetric version {asymmetric.__version__}",
    )

    # In order to allow the CLI utility grow, the parser will include an
    # initial argumment to determine which subparser will be executed.

    # Create subparsers
    subparsers = parser.add_subparsers(help="Action to be executed.")

    # Documentation parser
    generate_documentation_subparser(subparsers)

    return parser


def generate_documentation_subparser(subparsers: _SubParsersAction) -> ArgumentParser:
    """Generates the subparser for the auto-documentation option."""
    documentation_parser = subparsers.add_parser("docs")
    documentation_parser.set_defaults(action="docs")

    # Module name
    documentation_parser.add_argument(
        "module",
        metavar="module",
        help="Name of the module that uses the symmetric object.",
    )

    # Filename
    documentation_parser.add_argument(
        "-f",
        "--filename",
        dest="filename",
        default="openapi.json",
        help="Name of the file in where to write the OpenAPI documentation spec.",
    )

    return documentation_parser



"""
A module for every helper of asymmetric.
"""


def http_verb(dirty: str) -> str:
    """
    Given a 'dirty' HTTP verb (uppercased, with trailing whitespaces), strips
    it and returns it lowercased.
    """
    return dirty.strip().lower()


def humanize(module_name: str) -> str:
    """Transforms a module name into a pretty human-likable string."""
    module_name = module_name.lower()
    module_name = module_name.replace("_", " ").replace("-", " ")
    module_name = module_name.title()
    return module_name
